@@ -16,11 +16,6 @@
  * Boston, MA 021110-1307, USA.
  */
 
-#define _GNU_SOURCE
-#define _POSIX_C_SOURCE 200809
-#define _XOPEN_SOURCE 700
-#define _BSD_SOURCE
-
 #include "kerncompat.h"
 
 #include <unistd.h>
@@ -32,6 +27,7 @@
 #include <ftw.h>
 #include <wait.h>
 #include <assert.h>
+#include <getopt.h>
 
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -65,6 +61,7 @@ struct btrfs_receive
 	char *root_path;
 	char *dest_dir_path; /* relative to root_path */
 	char *full_subvol_path;
+	int dest_dir_chroot;
 
 	struct subvol_info *cur_subvol;
 
@@ -816,7 +813,8 @@ static struct btrfs_send_ops send_ops = {
 	.utimes = process_utimes,
 };
 
-static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd)
+static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd,
+		      u64 max_errors)
 {
 	int ret;
 	char *dest_dir_full_path;
@@ -840,9 +838,17 @@ static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd)
 
 	ret = find_mount_root(dest_dir_full_path, &r->root_path);
 	if (ret < 0) {
+		fprintf(stderr,
+			"ERROR: failed to determine mount point for %s: %s\n",
+			dest_dir_full_path, strerror(-ret));
+		ret = -EINVAL;
+		goto out;
+	}
+	if (ret > 0) {
+		fprintf(stderr,
+			"ERROR: %s doesn't belong to btrfs mount point\n",
+			dest_dir_full_path);
 		ret = -EINVAL;
-		fprintf(stderr, "ERROR: failed to determine mount point "
-			"for %s\n", dest_dir_full_path);
 		goto out;
 	}
 	r->mnt_fd = open(r->root_path, O_RDONLY | O_NOATIME);
@@ -853,14 +859,35 @@ static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd)
 		goto out;
 	}
 
-	/*
-	 * find_mount_root returns a root_path that is a subpath of
-	 * dest_dir_full_path. Now get the other part of root_path,
-	 * which is the destination dir relative to root_path.
-	 */
-	r->dest_dir_path = dest_dir_full_path + strlen(r->root_path);
-	while (r->dest_dir_path[0] == '/')
-		r->dest_dir_path++;
+	if (r->dest_dir_chroot) {
+		if (chroot(dest_dir_full_path)) {
+			ret = -errno;
+			fprintf(stderr,
+				"ERROR: failed to chroot to %s, %s\n",
+				dest_dir_full_path,
+				strerror(-ret));
+			goto out;
+		}
+		if (chdir("/")) {
+			ret = -errno;
+			fprintf(stderr,
+				"ERROR: failed to chdir to /, %s\n",
+				strerror(-ret));
+			goto out;
+		}
+		fprintf(stderr, "Chroot to %s\n", dest_dir_full_path);
+		r->root_path = strdup("/");
+		r->dest_dir_path = r->root_path;
+	} else {
+		/*
+		 * find_mount_root returns a root_path that is a subpath of
+		 * dest_dir_full_path. Now get the other part of root_path,
+		 * which is the destination dir relative to root_path.
+		 */
+		r->dest_dir_path = dest_dir_full_path + strlen(r->root_path);
+		while (r->dest_dir_path[0] == '/')
+			r->dest_dir_path++;
+	}
 
 	ret = subvol_uuid_search_init(r->mnt_fd, &r->sus);
 	if (ret < 0)
@@ -868,7 +895,8 @@ static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd)
 
 	while (!end) {
 		ret = btrfs_read_and_process_send_stream(r_fd, &send_ops, r,
-							 r->honor_end_cmd);
+							 r->honor_end_cmd,
+							 max_errors);
 		if (ret < 0)
 			goto out;
 		if (ret)
@@ -913,20 +941,31 @@ out:
 
 int cmd_receive(int argc, char **argv)
 {
-	int c;
 	char *tomnt = NULL;
 	char *fromfile = NULL;
 	struct btrfs_receive r;
 	int receive_fd = fileno(stdin);
-
+	u64 max_errors = 1;
 	int ret;
 
 	memset(&r, 0, sizeof(r));
 	r.mnt_fd = -1;
 	r.write_fd = -1;
 	r.dest_dir_fd = -1;
+	r.dest_dir_chroot = 0;
+
+	while (1) {
+		int c;
+		static const struct option long_opts[] = {
+			{ "max-errors", required_argument, NULL, 'E' },
+			{ "chroot", no_argument, NULL, 'C' },
+			{ NULL, 0, NULL, 0 }
+		};
+
+		c = getopt_long(argc, argv, "Cevf:", long_opts, NULL);
+		if (c < 0)
+			break;
 
-	while ((c = getopt(argc, argv, "evf:")) != -1) {
 		switch (c) {
 		case 'v':
 			g_verbose++;
@@ -937,6 +976,12 @@ int cmd_receive(int argc, char **argv)
 		case 'e':
 			r.honor_end_cmd = 1;
 			break;
+		case 'C':
+			r.dest_dir_chroot = 1;
+			break;
+		case 'E':
+			max_errors = arg_strtou64(optarg);
+			break;
 		case '?':
 		default:
 			fprintf(stderr, "ERROR: receive args invalid.\n");
@@ -957,15 +1002,15 @@ int cmd_receive(int argc, char **argv)
 		}
 	}
 
-	ret = do_receive(&r, tomnt, receive_fd);
+	ret = do_receive(&r, tomnt, receive_fd, max_errors);
 
 	return !!ret;
 }
 
 const char * const cmd_receive_usage[] = {
-	"btrfs receive [-ve] [-f <infile>] <mount>",
+	"btrfs receive [-ve] [-f <infile>] [--max-errors <N>] <mount>",
 	"Receive subvolumes from stdin.",
-	"Receives one or more subvolumes that were previously ",
+	"Receives one or more subvolumes that were previously",
 	"sent with btrfs send. The received subvolumes are stored",
 	"into <mount>.",
 	"btrfs receive will fail in case a receiving subvolume",
@@ -983,5 +1028,9 @@ const char * const cmd_receive_usage[] = {
 	"                 in the data stream. Without this option,",
 	"                 the receiver terminates only if an error",
 	"                 is recognized or on EOF.",
+	"-C|--chroot      confine the process to <mount> using chroot",
+	"--max-errors <N> Terminate as soon as N errors happened while",
+	"                 processing commands from the send stream.",
+	"                 Default value is 1. A value of 0 means no limit.",
 	NULL
 };
