@@ -44,11 +44,6 @@
 
 static u64 index_cnt = 2;
 
-#define DEFAULT_MKFS_FEATURES	(BTRFS_FEATURE_INCOMPAT_EXTENDED_IREF \
-		| BTRFS_FEATURE_INCOMPAT_SKINNY_METADATA)
-
-#define DEFAULT_MKFS_LEAF_SIZE 16384
-
 struct directory_name_entry {
 	char *dir_name;
 	char *path;
@@ -274,21 +269,21 @@ static void print_usage(void)
 {
 	fprintf(stderr, "usage: mkfs.btrfs [options] dev [ dev ... ]\n");
 	fprintf(stderr, "options:\n");
-	fprintf(stderr, "\t -A --alloc-start the offset to start the FS\n");
-	fprintf(stderr, "\t -b --byte-count total number of bytes in the FS\n");
-	fprintf(stderr, "\t -d --data data profile, raid0, raid1, raid5, raid6, raid10, dup or single\n");
-	fprintf(stderr, "\t -f --force force overwrite of existing filesystem\n");
-	fprintf(stderr, "\t -l --leafsize size of btree leaves\n");
-	fprintf(stderr, "\t -L --label set a label\n");
-	fprintf(stderr, "\t -m --metadata metadata profile, values like data profile\n");
-	fprintf(stderr, "\t -M --mixed mix metadata and data together\n");
-	fprintf(stderr, "\t -n --nodesize size of btree nodes\n");
-	fprintf(stderr, "\t -s --sectorsize min block allocation (may not mountable by current kernel)\n");
-	fprintf(stderr, "\t -r --rootdir the source directory\n");
-	fprintf(stderr, "\t -K --nodiscard do not perform whole device TRIM\n");
-	fprintf(stderr, "\t -O --features comma separated list of filesystem features\n");
-	fprintf(stderr, "\t -U --uuid specify the filesystem UUID\n");
-	fprintf(stderr, "\t -V --version print the mkfs.btrfs version and exit\n");
+	fprintf(stderr, "\t-A|--alloc-start START  the offset to start the FS\n");
+	fprintf(stderr, "\t-b|--byte-count SIZE    total number of bytes in the FS\n");
+	fprintf(stderr, "\t-d|--data PROFILE       data profile, raid0, raid1, raid5, raid6, raid10, dup or single\n");
+	fprintf(stderr, "\t-f|--force              force overwrite of existing filesystem\n");
+	fprintf(stderr, "\t-l|--leafsize SIZE      deprecated, alias for nodesize\n");
+	fprintf(stderr, "\t-L|--label LABEL        set a label\n");
+	fprintf(stderr, "\t-m|--metadata PROFILE   metadata profile, values like data profile\n");
+	fprintf(stderr, "\t-M|--mixed              mix metadata and data together\n");
+	fprintf(stderr, "\t-n|--nodesize SIZE      size of btree nodes\n");
+	fprintf(stderr, "\t-s|--sectorsize SIZE    min block allocation (may not mountable by current kernel)\n");
+	fprintf(stderr, "\t-r|--rootdir DIR        the source directory\n");
+	fprintf(stderr, "\t-K|--nodiscard          do not perform whole device TRIM\n");
+	fprintf(stderr, "\t-O|--features LIST      comma separated list of filesystem features\n");
+	fprintf(stderr, "\t-U|--uuid UUID          specify the filesystem UUID\n");
+	fprintf(stderr, "\t-V|--version            print the mkfs.btrfs version and exit\n");
 	fprintf(stderr, "%s\n", PACKAGE_STRING);
 	exit(1);
 }
@@ -1057,27 +1052,6 @@ static int zero_output_file(int out_fd, u64 size, u32 sectorsize)
 	return ret;
 }
 
-static int check_leaf_or_node_size(u32 size, u32 sectorsize)
-{
-	if (size < sectorsize) {
-		fprintf(stderr,
-			"Illegal leafsize (or nodesize) %u (smaller than %u)\n",
-			size, sectorsize);
-		return -1;
-	} else if (size > BTRFS_MAX_METADATA_BLOCKSIZE) {
-		fprintf(stderr,
-			"Illegal leafsize (or nodesize) %u (larger than %u)\n",
-			size, BTRFS_MAX_METADATA_BLOCKSIZE);
-		return -1;
-	} else if (size & (sectorsize - 1)) {
-		fprintf(stderr,
-			"Illegal leafsize (or nodesize) %u (not align to %u)\n",
-			size, sectorsize);
-		return -1;
-	}
-	return 0;
-}
-
 static int is_ssd(const char *file)
 {
 	blkid_probe probe;
@@ -1126,97 +1100,6 @@ static int is_ssd(const char *file)
 	return !atoi((const char *)&rotational);
 }
 
-#define BTRFS_FEATURE_LIST_ALL		(1ULL << 63)
-
-static const struct btrfs_fs_feature {
-	const char *name;
-	u64 flag;
-	const char *desc;
-} mkfs_features[] = {
-	{ "mixed-bg", BTRFS_FEATURE_INCOMPAT_MIXED_GROUPS,
-		"mixed data and metadata block groups" },
-	{ "extref", BTRFS_FEATURE_INCOMPAT_EXTENDED_IREF,
-		"increased hardlink limit per file to 65536" },
-	{ "raid56", BTRFS_FEATURE_INCOMPAT_RAID56,
-		"raid56 extended format" },
-	{ "skinny-metadata", BTRFS_FEATURE_INCOMPAT_SKINNY_METADATA,
-		"reduced-size metadata extent refs" },
-	{ "no-holes", BTRFS_FEATURE_INCOMPAT_NO_HOLES,
-		"no explicit hole extents for files" },
-	/* Keep this one last */
-	{ "list-all", BTRFS_FEATURE_LIST_ALL, NULL }
-};
-
-static void list_all_fs_features(void)
-{
-	int i;
-
-	fprintf(stderr, "Filesystem features available at mkfs time:\n");
-	for (i = 0; i < ARRAY_SIZE(mkfs_features) - 1; i++) {
-		char *is_default = "";
-
-		if (mkfs_features[i].flag & DEFAULT_MKFS_FEATURES)
-			is_default = ", default";
-		fprintf(stderr, "%-20s- %s (0x%llx%s)\n",
-				mkfs_features[i].name,
-				mkfs_features[i].desc,
-				mkfs_features[i].flag,
-				is_default);
-	}
-}
-
-static int parse_one_fs_feature(const char *name, u64 *flags)
-{
-	int i;
-	int found = 0;
-
-	for (i = 0; i < ARRAY_SIZE(mkfs_features); i++) {
-		if (name[0] == '^' &&
-			!strcmp(mkfs_features[i].name, name + 1)) {
-			*flags &= ~ mkfs_features[i].flag;
-			found = 1;
-		} else if (!strcmp(mkfs_features[i].name, name)) {
-			*flags |= mkfs_features[i].flag;
-			found = 1;
-		}
-	}
-
-	return !found;
-}
-
-static void process_fs_features(u64 flags)
-{
-	int i;
-
-	for (i = 0; i < ARRAY_SIZE(mkfs_features); i++) {
-		if (flags & mkfs_features[i].flag) {
-			printf("Turning ON incompat feature '%s': %s\n",
-				mkfs_features[i].name,
-				mkfs_features[i].desc);
-		}
-	}
-}
-
-
-/*
- * Return NULL if all features were parsed fine, otherwise return the name of
- * the first unparsed.
- */
-static char* parse_fs_features(char *namelist, u64 *flags)
-{
-	char *this_char;
-	char *save_ptr = NULL; /* Satisfy static checkers */
-
-	for (this_char = strtok_r(namelist, ",", &save_ptr);
-	     this_char != NULL;
-	     this_char = strtok_r(NULL, ",", &save_ptr)) {
-		if (parse_one_fs_feature(this_char, flags))
-			return this_char;
-	}
-
-	return NULL;
-}
-
 int main(int ac, char **av)
 {
 	char *file;
@@ -1230,16 +1113,16 @@ int main(int ac, char **av)
 	u64 alloc_start = 0;
 	u64 metadata_profile = 0;
 	u64 data_profile = 0;
-	u32 leafsize = max_t(u32, sysconf(_SC_PAGESIZE), DEFAULT_MKFS_LEAF_SIZE);
+	u32 nodesize = max_t(u32, sysconf(_SC_PAGESIZE),
+			BTRFS_MKFS_DEFAULT_NODE_SIZE);
 	u32 sectorsize = 4096;
-	u32 nodesize = leafsize;
 	u32 stripesize = 4096;
 	int zero_end = 1;
 	int fd;
 	int ret;
 	int i;
 	int mixed = 0;
-	int leaf_forced = 0;
+	int nodesize_forced = 0;
 	int data_profile_opt = 0;
 	int metadata_profile_opt = 0;
 	int discard = 1;
@@ -1255,32 +1138,31 @@ int main(int ac, char **av)
 	int saved_optind;
 	char estr[100];
 	char *fs_uuid = NULL;
-	u64 features = DEFAULT_MKFS_FEATURES;
+	u64 features = BTRFS_MKFS_DEFAULT_FEATURES;
 
 	while(1) {
 		int c;
-		int option_index = 0;
 		static const struct option long_options[] = {
-			{ "alloc-start", 1, NULL, 'A'},
-			{ "byte-count", 1, NULL, 'b' },
-			{ "force", 0, NULL, 'f' },
-			{ "leafsize", 1, NULL, 'l' },
-			{ "label", 1, NULL, 'L'},
-			{ "metadata", 1, NULL, 'm' },
-			{ "mixed", 0, NULL, 'M' },
-			{ "nodesize", 1, NULL, 'n' },
-			{ "sectorsize", 1, NULL, 's' },
-			{ "data", 1, NULL, 'd' },
-			{ "version", 0, NULL, 'V' },
-			{ "rootdir", 1, NULL, 'r' },
-			{ "nodiscard", 0, NULL, 'K' },
-			{ "features", 1, NULL, 'O' },
+			{ "alloc-start", required_argument, NULL, 'A'},
+			{ "byte-count", required_argument, NULL, 'b' },
+			{ "force", no_argument, NULL, 'f' },
+			{ "leafsize", required_argument, NULL, 'l' },
+			{ "label", required_argument, NULL, 'L'},
+			{ "metadata", required_argument, NULL, 'm' },
+			{ "mixed", no_argument, NULL, 'M' },
+			{ "nodesize", required_argument, NULL, 'n' },
+			{ "sectorsize", required_argument, NULL, 's' },
+			{ "data", required_argument, NULL, 'd' },
+			{ "version", no_argument, NULL, 'V' },
+			{ "rootdir", required_argument, NULL, 'r' },
+			{ "nodiscard", no_argument, NULL, 'K' },
+			{ "features", required_argument, NULL, 'O' },
 			{ "uuid", required_argument, NULL, 'U' },
 			{ NULL, 0, NULL, 0}
 		};
 
 		c = getopt_long(ac, av, "A:b:fl:n:s:m:d:L:O:r:U:VMK",
-				long_options, &option_index);
+				long_options, NULL);
 		if (c < 0)
 			break;
 		switch(c) {
@@ -1295,10 +1177,11 @@ int main(int ac, char **av)
 				data_profile_opt = 1;
 				break;
 			case 'l':
+				fprintf(stderr,
+			"WARNING: --leafsize is deprecated, use --nodesize\n");
 			case 'n':
 				nodesize = parse_size(optarg);
-				leafsize = parse_size(optarg);
-				leaf_forced = 1;
+				nodesize_forced = 1;
 				break;
 			case 'L':
 				label = parse_label(optarg);
@@ -1314,7 +1197,7 @@ int main(int ac, char **av)
 				char *orig = strdup(optarg);
 				char *tmp = orig;
 
-				tmp = parse_fs_features(tmp, &features);
+				tmp = btrfs_parse_fs_features(tmp, &features);
 				if (tmp) {
 					fprintf(stderr,
 						"Unrecognized filesystem feature '%s'\n",
@@ -1324,7 +1207,7 @@ int main(int ac, char **av)
 				}
 				free(orig);
 				if (features & BTRFS_FEATURE_LIST_ALL) {
-					list_all_fs_features();
+					btrfs_list_all_fs_features(0);
 					exit(0);
 				}
 				break;
@@ -1359,9 +1242,7 @@ int main(int ac, char **av)
 		}
 	}
 	sectorsize = max(sectorsize, (u32)sysconf(_SC_PAGESIZE));
-	if (check_leaf_or_node_size(leafsize, sectorsize))
-		exit(1);
-	if (check_leaf_or_node_size(nodesize, sectorsize))
+	if (btrfs_check_nodesize(nodesize, sectorsize))
 		exit(1);
 	saved_optind = optind;
 	dev_cnt = ac - optind;
@@ -1427,7 +1308,7 @@ int main(int ac, char **av)
 				BTRFS_BLOCK_GROUP_RAID0 : 0; /* raid0 or single */
 		}
 	} else {
-		u32 best_leafsize = max_t(u32, sysconf(_SC_PAGESIZE), sectorsize);
+		u32 best_nodesize = max_t(u32, sysconf(_SC_PAGESIZE), sectorsize);
 
 		if (metadata_profile_opt || data_profile_opt) {
 			if (metadata_profile != data_profile) {
@@ -1437,34 +1318,33 @@ int main(int ac, char **av)
 			}
 		}
 
-		if (!leaf_forced) {
-			leafsize = best_leafsize;
-			nodesize = best_leafsize;
-			if (check_leaf_or_node_size(leafsize, sectorsize))
+		if (!nodesize_forced) {
+			nodesize = best_nodesize;
+			if (btrfs_check_nodesize(nodesize, sectorsize))
 				exit(1);
 		}
-		if (leafsize != sectorsize) {
+		if (nodesize != sectorsize) {
 			fprintf(stderr, "Error: mixed metadata/data block groups "
 				"require metadata blocksizes equal to the sectorsize\n");
 			exit(1);
 		}
 	}
 
-	/* Check device/block_count after the leafsize is determined */
-	if (block_count && block_count < btrfs_min_dev_size(leafsize)) {
+	/* Check device/block_count after the nodesize is determined */
+	if (block_count && block_count < btrfs_min_dev_size(nodesize)) {
 		fprintf(stderr,
 			"Size '%llu' is too small to make a usable filesystem\n",
 			block_count);
 		fprintf(stderr,
 			"Minimum size for btrfs filesystem is %llu\n",
-			btrfs_min_dev_size(leafsize));
+			btrfs_min_dev_size(nodesize));
 		exit(1);
 	}
 	for (i = saved_optind; i < saved_optind + dev_cnt; i++) {
 		char *path;
 
 		path = av[i];
-		ret = test_minimum_size(path, leafsize);
+		ret = test_minimum_size(path, nodesize);
 		if (ret < 0) {
 			fprintf(stderr, "Failed to check size for '%s': %s\n",
 				path, strerror(-ret));
@@ -1476,7 +1356,7 @@ int main(int ac, char **av)
 				path);
 			fprintf(stderr,
 				"Minimum size for each btrfs device is %llu.\n",
-				btrfs_min_dev_size(leafsize));
+				btrfs_min_dev_size(nodesize));
 			exit(1);
 		}
 	}
@@ -1546,7 +1426,7 @@ int main(int ac, char **av)
 	blocks[0] = BTRFS_SUPER_INFO_OFFSET;
 	for (i = 1; i < 7; i++) {
 		blocks[i] = BTRFS_SUPER_INFO_OFFSET + 1024 * 1024 +
-			leafsize * i;
+			nodesize * i;
 	}
 
 	/*
@@ -1561,11 +1441,10 @@ int main(int ac, char **av)
 		features |= BTRFS_FEATURE_INCOMPAT_RAID56;
 	}
 
-	process_fs_features(features);
+	btrfs_process_fs_features(features);
 
 	ret = make_btrfs(fd, file, label, fs_uuid, blocks, dev_block_count,
-			 nodesize, leafsize,
-			 sectorsize, stripesize, features);
+			 nodesize, sectorsize, stripesize, features);
 	if (ret) {
 		fprintf(stderr, "error during mkfs: %s\n", strerror(-ret));
 		exit(1);
@@ -1643,7 +1522,7 @@ raid_groups:
 
 	printf("fs created label %s on %s\n\tnodesize %u leafsize %u "
 	    "sectorsize %u size %s\n",
-	    label, first_file, nodesize, leafsize, sectorsize,
+	    label, first_file, nodesize, nodesize, sectorsize,
 	    pretty_size(btrfs_super_total_bytes(root->fs_info->super_copy)));
 
 	btrfs_commit_transaction(trans, root);
